@@ -28,8 +28,25 @@ class ErplyAuth(object):
 
 class Erply(object):
 
-    ERPLY_GET = ('getCustomerGroups', 'getCustomers', 'getDocuments', \
-                 'getProducts', 'getSalesDocuments', 'verifyUser')
+    ERPLY_GET = (
+        # TODO: This list is still incomplete
+         'getAddresses'
+        ,'getAddressTypes'
+        ,'getCustomers'
+        ,'getCustomerGroups'
+        ,'getProducts'
+        ,'getProductCategories'
+        ,'getProductCostForSpecificAmount'     # untested
+        ,'getProductGroups'
+        ,'getProductPrices'                    # untested, broken ??
+        ,'getProductPriorityGroups'            # untested
+        ,'getProductStock'                     # untested
+        ,'getProductUnits'
+        ,'getSalesDocuments'
+        ,'getServices'
+        # ,'getDocuments'       Unimplemented from ERPLY side :(
+        ,'verifyUser'
+    )
     ERPLY_CSV = ('getProductStockCSV',)
     ERPLY_POST = ('saveProduct',)
 
@@ -72,54 +89,91 @@ class Erply(object):
         return ErplyCSVResponse(self, requests.post(self.api_url, data=data, headers=self.headers))
 
     def handle_get(self, request, _page=None, _per_page=None, _response=None, *args, **kwargs):
-        data = dict(request=request)
-        data.update(self.payload if request != 'verifyUser' else self._payload)
-        data.update(kwargs)
+        _is_bulk = kwargs.pop('_is_bulk', False)
+        data = kwargs.copy()
         if _page:
             data['pageNo'] = _page + 1
         if _per_page:
             data['recordsOnPage'] = _per_page
+        if _is_bulk:
+            data.update(requestName=request)
+            return data
+        data.update(request=request)
+        data.update(self.payload if request != 'verifyUser' else self._payload)
         r = requests.post(self.api_url, data=data, headers=self.headers)
         if _response:
             _response.update(r, _page)
-        return ErplyResponse(self, r, request, _page)
+        return ErplyResponse(self, r, request, _page, *args, **kwargs)
 
     def handle_post(self, request, *args, **kwargs):
-        data = dict(request=request)
+        _is_bulk = kwargs.pop('_is_bulk', False)
+        data = kwargs.copy()
+        if _is_bulk:
+            data.update(requestName=request)
+            return data
+        data.update(request=request)
         data.update(self.payload)
-        data.update(**kwargs)
         r = requests.post(self.api_url, data=data, headers=self.headers)
-        return ErplyResponse(self, r, request)
+        return ErplyResponse(self, r, request, *args, **kwargs)
+
+    def handle_bulk(self, _requests):
+        data = self.payload
+        data.update(requests=_requests)
+        return ErplyBulkResponse(self, requests.post(self.api_url, data=data))
 
     def __getattr__(self, attr):
         _attr = None
+        _is_bulk = len(attr) > 5 and attr.endswith('_bulk')
+        if _is_bulk:
+            attr = attr[:-5]
         if attr in self.ERPLY_GET:
             def method(*args, **kwargs):
                 _page = kwargs.get('_page', 0)
                 _response = kwargs.get('_response', None)
-                return self.handle_get(attr, _page, _response, *args, **kwargs)
+                return self.handle_get(attr, _page, _response, _is_bulk=_is_bulk, *args, **kwargs)
+            _attr = method
+        elif attr in self.ERPLY_POST:
+            def method(*args, **kwargs):
+                return self.handle_post(attr, _is_bulk=_is_bulk, *args, **kwargs)
             _attr = method
         elif attr in self.ERPLY_CSV:
             def method(*args, **kwargs):
                 return self.handle_csv(attr.replace('CSV', ''), *args, **kwargs)
             _attr = method
-        elif attr in self.ERPLY_POST:
-            def method(*args, **kwargs):
-                return self.handle_post(attr, *args, **kwargs)
-            _attr = method
         if _attr:
             self.__dict__[attr] = _attr
             return _attr
         raise AttributeError
 
 
+class ErplyBulkRequest(object):
+    def __init__(self, erply,  _json_dumps):
+        self.calls = []
+        self.erply = erply
+        self.json_dumper = _json_dumps
+
+    def attach(self, attr, *args, **kwargs):
+        if attr in self.erply.ERPLY_GET or attr in self.erply.ERPLY_POST:
+            self.calls.append((getattr(self.erply, '{}_bulk'.format(attr)), args, kwargs))
+
+    def __call__(self,):
+        _requests = []
+        for n, request in enumerate(self.calls, start=1):
+            _call, _args, _kwargs = request
+            _kwargs.update(requestID=n)
+            _requests.append(_call(*_args, **_kwargs))
+        return self.erply.handle_bulk(self.json_dumper(_requests))
+
+
 class ErplyResponse(object):
 
-    def __init__(self, erply, response, request, page=0):
+    def __init__(self, erply, response, request, page=0, *args, **kwargs):
         self.request = request
         self.erply = erply
         self.error = None
 
+        self.kwargs = kwargs
+
         if response.status_code != requests.codes.ok:
             print ('Request failed with error code {}'.format(response.status_code))
             raise ValueError
@@ -155,12 +209,12 @@ class ErplyResponse(object):
         raise ValueError
 
     def fetch_records(self, page):
-        self.erply.handle_get(self.request, _page=page, _per_page=self.per_page, _response=self)
+        self.erply.handle_get(self.request, _page=page, _per_page=self.per_page, _response=self, **self.kwargs)
 
     def update(self, data, page):
         items = data.json().get('records')
-        assert len(items) != 0
-        self.records[page] = items
+        if len(items):
+            self.records[page] = items
 
     def __getitem__(self, key):
         if isinstance(key, slice):
@@ -200,3 +254,34 @@ class ErplyCSVResponse(object):
             reader = csv.reader(f.text.splitlines())
             reader.next()
             return reader
+
+
+class ErplyBulkResponse(object):
+    def __init__(self, erply, response):
+        if response.status_code != requests.codes.ok:
+            print ('Request failed with error code {}'.format(response.status_code))
+            raise ValueError
+
+        self.data = response.json()
+        status = self.data.get('status', {})
+        if not status:
+            print ("Malformed response")
+            raise ValueError
+
+        self.error = status.get('errorCode')
+        self._requests = self.data.get('requests')
+
+
+    @property
+    def records(self):
+        if self._requests is None:
+            raise ValueError
+        for el in self._requests:
+            _status = el.get('status')
+            if _status.get('responseStatus') == 'error':
+                print ('Request failed: requestID: {} errorField: {}'.format(
+                        _status.get('requestID'),
+                        _status.get('errorField'),
+                       ))
+            else:
+                yield el.get('records')
